/* -------------------------------------------------------------------------------
 * DFM-1a Digital Filter Module
 * Copyright (C) 2010, 2011
 *
 * Tony Hardie-Bick <tony@entitysynth.net> - Java version
 * Jonny Stutters <jstutters@jeremah.co.uk> - SuperCollider port
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * ------------------------------------------------------------------------------ */

#include <cmath>
#include "Dfm1aProcessing.h"

/**
 * Filters audio samples.
 *
 * The arguments define the target values which will have been reached by the
 * filter parameters when the method returns. If a large number of samples
 * (more than a thousand) needs to be processed, this method should be called
 * multiple times with a smaller number of samples. The audio data has a
 * typical range from minus to plus one, but can be greater. The input samples
 * are multiplied by the input gain so that high distortion levels can be
 * readily achieved with ordinary signals. Output samples are always in the
 * range minus to plus one. Filter self-oscillation also covers this range.
 *
 * @param inputLevel Input level (0..n, typically 0.5..5.0)
 * @param frequency Resonant frequency in hertz.
 * @param resonance Resonance level (1 = self osc., and can be higher).
 * @param filterType Filter type (below 0.5 gives low pass, otherwise high).
 * @param noiseLevel Level of noise in the filter (0..n, typically 0.0003)
 * @param inBuffer Input audio samples
 * @param outBuffer Output audio samples, will contain filtered audio on exit
 * @param modBuffer Audio rate frequency modulation samples
 * @param nSamples Number of audio samples to process.
 * @param sampleRate Sample rate in hertz.
 * @param df Dfm1 state variables
 */

void Dfm1aProcessing::filter ( 
    
    float inputLevel,
    float frequency,
    float resonance,
    float filterType,
    float noiseLevel,
    float inBuffer [],
    float outBuffer [],
    float modBuffer [],
    int nSamples,
    float sampleRate,
    Dfm1aState *df ) {

  if (nSamples < 1) return;

  double minFreq = Dfm1Lut::getMinFreq(sampleRate);
  double maxFreq = Dfm1Lut::getMaxFreq(sampleRate);

  if(frequency < minFreq) frequency = (float) minFreq;
  else if(frequency > maxFreq) frequency = (float) maxFreq;

  if (resonance < 0.0f) resonance = 0.0f;
  if (resonance > 10.0f) resonance = 10.0f;

  double coefficients[3];
  
  /* Use different buffer for input and output */

  float *input = inBuffer;
  float *output = outBuffer;
  
  /* Copy state into local variables for speed */

  double f = df->f; // Frequency
  double p = df->p; // LPF/HPF selector 
  double q = df->q; // Resonance 
  double s = df->s; // Noise level

  double za = df->za; // First stage integrator
  double zb = df->zb; // Second stage integrator
  double zh = df->zh; // High pass differentiator
  double zr = df->zr; // Resonance differentiator
  double zy = df->zy; // Filter output 
  
  NoiseGenState* ng = &(df->ng); // Noise generator state (pointer)
  
  double rcfs = 1.0 / sampleRate; // For speed (avoids an audio rate division)

  /* Obtain target values ("rt" means "target value for parameter r") */

  double ft = frequency; // Frequency target value
  double qt = resonance; // Resonance   "     
  double pt = filterType >= 0.5 ? 1.0 : 0.0; // LPF/HPF selector

  double st = noiseLevel * NOISE_GAIN; 
  if(st < NOISE) st = NOISE; // Set minimum noise level

  /* Calculate parameter interpolation increments */

  double i = 1.0 / nSamples;

  double fi = checkValue((ft - f) * i); 
  double pi = checkValue((pt - p) * i); 
  double qi = checkValue((qt - q) * i); 
  double si = checkValue((st - s) * i); 

  /* Sample loop */
  
  for(int t = 0; t < nSamples; t++) {
    
    f += fi; p += pi; q += qi; s += si; // Increment parameters

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
     * Frequency Modulation
     *
     * Modulating samples in the range -1 to +1 modulate the filter to a
     * depth approximately equal to the cutoff frequency. In other words, a
     * modulating sample value of -1 would set the instantaneous filter
     * frequency to 0Hz. However the filter cutoff cannot be set below its
     * minimum of around 1Hz. If the modulation is simply limited, a
     * discontinuity arises, resulting in digital artefacts.  Instead, the
     * instantaneous modulation depth is limited smoothly as it approaches
     * the filter's natural minimum or maximum frequency limits, with a
     * transfer function gradient of zero coinciding with the filter's
     * minimum or maximum frequency. As a consequence, the modulation is
     * non-linear, asymmetric and stronger at low sample amplitudes.
     * Modulation noise is added to simulate the effect of analogue
     * component noise in the modulation process.
     * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    double fm = modBuffer[t] + s * NoiseGen::rnor(ng); // Add modulation noise
    fm *= f; // Modulation depth proportional to resonant frequency

    double range; // Frequency range for current modulation sample
    if(fm < 0.0) { range = f - minFreq; } // Negative FM
    else { range = maxFreq - f; } // Positive FM
    
    if(range > 0.1) { // Only modulate when available range is more than 0.1Hz

      fm /= range; // Normalise to limit of curved transfer function

      /* Perform fixed limiting so curve doesn't go back on itself */
      fm = fm < -1.0 ? - 1.0 : fm > 1.0 ? 1.0 : fm;

      /* Apply curved transfer function */
      double nx = N * fm;
      fm = fm - 2.0 * nx * nx * nx;
        
      fm *= 3.0 / 2.0; // Amplify so maximum is at unity
      fm *= range; // Restore original range

      fm += f; // Modulate frequency

    } else { fm = f; }

    Dfm1Lut::getCoefficients(fm * rcfs, sampleRate, coefficients);

    double a = coefficients[0]; // Feedback for first integrator
    double b = coefficients[1]; // Feedback for second integrator
    double r = coefficients[2] * q; // Apply resonance correction factor

    double l = inputLevel * PRE_GAIN * (1.0 - a); // Low pass gain conflated with first integrator gain
    double h = inputLevel * PRE_GAIN; // High pass gain

    /* Crossfade between high pass and low pass input */

    l *= (1.0 - p); 
    h *= p; 

    double x = input[t];

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
     * DFM-1 Filter 
     *
     *                    . - a - .                . - b - .
     *                    |       |                |       |
     *                    |       z' (za)          |       z' (zb)
     *                    |       |                |       |
     *      x[t]          v       |                v       |  q[t]               y[t]
     * lpf --->-- l -->-- + --------->-- 1-b -->-- + ---------->--- f(q[t]) -- + ---->
     *                  / ^                      / ^                         / |
     *                s   |                    s   |                       s   |
     *              /     r                  /    + -                    /     |
     *        noise       |            noise     /   \             noise       |
     *                   - +                    |     z' (zh)                  |
     *                  /   \                   |     |                        |
     *                 |     |    hpf -- h -->--------'                        |
     *           (zr)  z'    |                                                 |
     *                 |     |   r[t]              (zy)                        |
     *                 `-----------<--------------- z' ------------<-----------'
     *
     *
     * The filter's frequency is defined by a and b. These take different values,
     * simulating different capacitor values in each of the two stages. Resonance is
     * defined by r. The lpf input gain is provided by l (conflated with 1-a to prescale
     * the first stage). The hpf input gain is provided by h. Distortion is provided by
     * f(q[t]), in which the sample magnitude is limited to unity, and then the transfer
     * function f(x) = x - 2(nx)^3 is applied. The value of n is set so the maximum
     * input values coincide with the function's maxima, and the filter's output is
     * multiplied by a scaling factor to compensate for the resulting reduction in
     * amplitude of loud signals.
     *
     * The noise has two purposes: Disruption and dithering. In an analogue circuit,
     * a large number of noise sources is accumulated, resulting in rare high amplitude
     * events that can push the filter further into or out of distortion, knocking the
     * filter's resonant phase-locked loop out of or into phase with a signal harmonic.
     * Noise is applied so the magnitude of its low frequency components increases 
     * inversely with respect to filter cutoff frequency.
     *
     * Dithering ensures that amplitude quantization artefacts do not accumulate to
     * audible levels as a result of integration. In a floating point implementation, 
     * dither also ensures that denormalisation events are rare.
     * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  
    za = x * l + r * (zy - zr) + a * za + s * NoiseGen::rnor(ng);
    zb = za * (1.0 - b) + (x - zh) * h + b * zb + s * NoiseGen::rnor(ng);
    zh = x;
    zr = zy;
    
    zb = zb < -1.0 ? -1.0 : zb > 1.0 ? 1.0 : zb;
    double nx = N * zb;
    zy = zb - 2.0 * nx * nx * nx;
    
    zy += s * NoiseGen::rnor(ng);
    
    output[t]  = (float)(zy * POST_GAIN);

  }

  /* After loop, store filter state */
  
  df->f = ft; df->p = pt; df->q = qt; df->s = st; 
  df->za = za; df->zb = zb; df->zh = zh; df->zr = zr; df->zy = zy;

}

/**
 * Checks for parameter denormalisation.
 *
 * This method anticipates any denormalisation errors which may occur if a
 * parameter increment is extremely small, returning zero if this is the
 * case.    
 *
 * @param pi Parameter increment
 * @return Corrected parameter increment
 */

double Dfm1aProcessing::checkValue(double pi) {

  /* This check is very simple, and assumes that less than a thousand 
   * samples are processed each call, and that the audio sample rate is 
   * no more than a few hundred kilohertz. */

  const double DENORM_THRESHOLD = 0.000001;
  return std::abs(pi) < DENORM_THRESHOLD ? 0.0 : pi;

}

/**
 * Initializes the filter.
 *
 * Sets the filter state variables and current parameter values to zero.
 */

void Dfm1aProcessing::initialize(Dfm1aState* df) {

  df->f = df->p = df->q = df->s = 0;
  df->za = df->zb = df->zr = df->zy = 0;
  
  NoiseGen::initialize(&(df->ng));

}
